from concurrent.futures import Future, wait
from distutils.util import strtobool
import logging
import os
import sys
import tempfile

from iopipe.compat import StringIO, PY37
from iopipe.plugins import Plugin
from iopipe.signer import get_signed_request

from .formatter import JSONFormatter
from .request import upload_log_data
from .stream import StreamToLogger
from .wrapper import LogWrapper


class LoggerPlugin(Plugin):
    name = "logger"
    version = "0.1.0"
    homepage = "https://github.com/iopipe/iopipe-python#logger-plugin"

    def __init__(
        self,
        name=None,
        level=logging.INFO,
        enabled=False,
        redirect_stdout=True,
        use_tmp=False,
    ):
        """
        Instantiates the logger plugin

        :param name: Specify custom log name.
        :type name: str
        :param level: Specify a log level for the handler.
        :type level: int
        :param enabled: Whether or not to enable the plugin.
        :type enabled: bool
        :param redirect_stdout: Whether or not to redirect stdout.
        :type redirect_print: bool
        :param use_tmp: Write logs to the /tmp directory instead of a memory buffer.
        :type use_tmp: bool
        """
        self._enabled = enabled
        self.handler = None
        self.level = level
        self.logger = logging.getLogger(name)
        self.redirect_stdout = redirect_stdout
        self.use_tmp = use_tmp

        # Due to a change in python3.7 runtime's logging config, we cannot redirect
        # stdout without resulting in a recursion error.
        if PY37:
            self.redirect_stdout = False

        if self.enabled:
            self.init_handler()

    @property
    def enabled(self):
        return self._enabled is True or bool(
            strtobool(os.getenv("IOPIPE_LOGGER_ENABLED", "false"))
        )

    def init_handler(self):
        formatter = JSONFormatter()

        self.handler = logging.StreamHandler(StringIO())
        self.handler.setFormatter(formatter)
        self.handler.setLevel(self.level)
        self.logger.addHandler(self.handler)
        self.logger.setLevel(self.level)

    def pre_setup(self, iopipe):
        self.iopipe = iopipe

    def post_setup(self, iopipe):
        if self.enabled and iopipe.config["debug"] is True:
            self.handler.setLevel(logging.DEBUG)
            self.logger.setLevel(logging.DEBUG)

    def pre_invoke(self, event, context):
        self.context = context
        self.signed_request = None

        if self.enabled:
            self.context.iopipe.register(
                "log", LogWrapper(self.logger, context), force=True
            )

            self.signed_request = self.iopipe.submit_future(
                get_signed_request, self.iopipe.config, self.context, ".log"
            )

            if not self.handler:
                self.init_handler()

            if self.use_tmp is True:
                self.handler.stream = tempfile.NamedTemporaryFile(
                    delete=False, mode="w"
                )
            else:
                self.handler.stream = StringIO()

            if self.redirect_stdout is True:
                sys.stdout = StreamToLogger(self.logger)

    def post_invoke(self, event, context):
        if self.enabled and self.redirect_stdout is True:
            sys.stdout = sys.__stdout__

    def post_response(self, response):
        pass

    def pre_report(self, report):
        if self.enabled:
            self.handler.flush()

            if self.handler.stream.tell():
                stream = self.handler.stream

                if hasattr(stream, "getvalue"):
                    stream = StringIO(stream.getvalue())

                if hasattr(stream, "file"):
                    stream = stream.name
                    self.handler.stream.close()
                if self.signed_request is not None:

                    if isinstance(self.signed_request, Future):
                        wait([self.signed_request])
                        self.signed_request = self.signed_request.result()

                if (
                    self.signed_request is not None
                    and "signedRequest" in self.signed_request
                ):
                    self.iopipe.submit_future(
                        upload_log_data,
                        self.signed_request["signedRequest"],
                        stream,
                        self.iopipe.config,
                    )
                    if "jwtAccess" in self.signed_request:
                        plugin = next(
                            (p for p in report.plugins if p["name"] == self.name)
                        )

                        if "uploads" not in plugin:
                            plugin["uploads"] = []
                        plugin["uploads"].append(self.signed_request["jwtAccess"])

    def post_report(self, report):
        pass



import decimal
import logging
import numbers
import uuid
import warnings

from . import constants
from .compat import string_types


class LogWrapper(object):
    def __init__(self, context):
        self.context = context
        self.logger = logging.getLogger()

    def __call__(self, key, value):
        warnings.warn(
            "Calling context.iopipe.log() has been deprecated, use "
            "context.iopipe.metric() instead"
        )

        self.context.metric(key, value)

    def __getattr__(self, name):
        return getattr(self.logger, name)


class ContextWrapper(object):
    def __init__(self, base_context, instance, **kwargs):
        self.base_context = base_context
        self.instance = instance
        self.iopipe = IOpipeContext(self.instance, **kwargs)

    def __getattr__(self, name):
        return getattr(self.base_context, name)


class IOpipeContext(object):
    def __init__(self, instance, **kwargs):
        self.instance = instance
        self.log = LogWrapper(self)
        self.disabled = False
        self.is_step_function = kwargs.pop("step_function", False)
        self.step_meta = None

    def metric(self, key, value):
        if self.instance.report is None:
            warnings.warn(
                "Attempting to add metrics before function decorated with IOpipe. "
                "This metric will not be recorded."
            )
            return

        name = str(key)
        if len(name) > constants.METRIC_NAME_LIMIT:
            warnings.warn(
                "Custom metric of name %s is longer than allowed limit of "
                "%s characters. "
                "This metric will not be recorded."
                % (name, constants.METRIC_NAME_LIMIT)
            )
            return

        event = {"name": name}

        # Add numerical values to report
        # We typecast decimals as strings: not JSON serializable and casting to floats
        # can result in rounding errors.
        if isinstance(value, numbers.Number) and not isinstance(value, decimal.Decimal):
            event["n"] = value
        else:
            event["s"] = str(value)

        self.instance.report.custom_metrics.append(event)

        if not name.startswith("@iopipe"):
            self.label("@iopipe/metrics")

    def label(self, *names):
        if self.instance.report is None:
            warnings.warn(
                "Attempting to add label before function decorated with IOpipe. "
                "This label will not be recorded."
            )
            return

        for name in names:
            if not isinstance(name, string_types):
                warnings.warn(
                    "Attempted to add a label that is not of type string. This label "
                    "will not be recorded."
                )
                continue

            if len(name) > constants.METRIC_NAME_LIMIT:
                warnings.warn(
                    "Label of name %s is longer than allowed limit of %s characters. "
                    "This label will not be recorded."
                    % (name, constants.METRIC_NAME_LIMIT)
                )
                continue

            self.instance.report.labels.add(name)

    def error(self, error):
        if self.instance.report is None:
            warnings.warn(
                "An exception occurred before function was decorated with IOpipe. "
                "This exception will not be recorded."
            )
            raise error

        self.instance.report.prepare(error)
        self.instance.run_hooks("pre:report")
        self.instance.report.send()
        self.instance.run_hooks("post:report")
        raise error

    def event_type(self, event_Type):
        if self.instance.report is None:
            return

        self.instance.report.report["eventType"] = event_Type

    def register(self, name, value, force=False):
        if not hasattr(self, name) or force:
            setattr(self, name, value)

    def unregister(self, name):
        if hasattr(self, name):
            delattr(self, name)

    def disable(self):
        self.disabled = True

    def collect_step_meta(self, event):
        if self.is_step_function:
            self.step_meta = event.get("iopipe", {"id": str(uuid.uuid4()), "step": 0})

    def inject_step_meta(self, response):
        if self.step_meta and isinstance(response, dict):
            step_meta = self.step_meta.copy()
            step_meta["step"] += 1
            response["iopipe"] = step_meta



from distutils.util import strtobool
import os
import warnings

from .collector import get_collector_path, get_hostname


def set_config(**config):
    """
    Returns IOpipe configuration options, setting defaults as necessary.
    """
    config.setdefault("debug", bool(strtobool(os.getenv("IOPIPE_DEBUG", "false"))))
    config.setdefault("enabled", bool(strtobool(os.getenv("IOPIPE_ENABLED", "true"))))
    config.setdefault("host", get_hostname())
    config.setdefault("install_method", os.getenv("IOPIPE_INSTALL_METHOD", "manual"))
    config.setdefault("network_timeout", os.getenv("IOPIPE_NETWORK_TIMEOUT", 5000))
    config.setdefault("path", get_collector_path())
    config.setdefault("plugins", [])
    config.setdefault("sync_http", False)
    config.setdefault("timeout_window", os.getenv("IOPIPE_TIMEOUT_WINDOW", 150))
    config.setdefault(
        "token", os.getenv("IOPIPE_TOKEN") or os.getenv("IOPIPE_CLIENTID") or ""
    )

    if "client_id" in config:
        config["token"] = config.pop("client_id")

    if "url" in config:
        url = config.pop("url")

        config["host"] = get_hostname(url)
        config["path"] = get_collector_path(url)

    if "." in str(config["network_timeout"]):
        warnings.warn(
            "IOpipe's 'network_timeout' is now in milliseconds, expressed as an integer"
        )

    try:
        config["debug"] = bool(config["debug"])
    except ValueError:
        config["debug"] = False

    try:
        config["network_timeout"] = int(config["network_timeout"]) / 1000.0
    except ValueError:
        config["network_timeout"] = 5.0

    if "." in str(config["timeout_window"]):
        warnings.warn(
            "IOpipe's 'timeout_window' is now in milliseconds, expressed as an integer"
        )

    try:
        config["timeout_window"] = int(config["timeout_window"]) / 1000.0
    except ValueError:
        config["timeout_window"] = 0.15

    return config



from .util import get_value


class ResponseType(object):
    keys = []

    def __init__(self, event_type):
        self.event_type = event_type

    def collect(self, response):
        response_info = {}

        for key in self.keys:
            if isinstance(key, tuple):
                old_key, new_key = key
            else:
                old_key = new_key = key
            value = get_value(response, old_key)
            if value is not None:
                response_info[
                    "@iopipe/event-info.%s.response.%s" % (self.event_type, new_key)
                ] = value

        return response_info


class LambdaProxy(ResponseType):
    keys = ["statusCode"]



import collections
import uuid
import wrapt

from .util import ensure_utf8

Request = collections.namedtuple(
    "Request", ["command", "key", "hostname", "port", "connectionName", "db", "table"]
)


def collect_pymongo_metrics(context, trace, instance, response):
    from pymongo.cursor import Cursor
    from pymongo.results import (
        BulkWriteResult,
        DeleteResult,
        InsertOneResult,
        InsertManyResult,
        UpdateResult,
    )

    command, key = None, None

    for command_name, key_attr, result_type in [
        ("bulk_write", "inserted_count", BulkWriteResult),
        ("delete", "deleted_count", DeleteResult),
        ("insert_one", "inserted_id", InsertOneResult),
        ("insert_many", "inserted_ids", InsertManyResult),
        ("update", ("upserted_id", "modified_count"), UpdateResult),
    ]:
        if isinstance(response, result_type):
            command = command_name
            if isinstance(key_attr, tuple):
                for attr in key_attr:
                    key = getattr(response, attr, None)
                    if key is not None:
                        break
            else:
                key = getattr(response, key_attr, None)
            if key is not None:
                key = str(key)
            break

    if isinstance(response, Cursor):  # pragman: no cover
        command = "find"
        key = response.retrieved

    hostname, port = None, None

    database = getattr(instance, "database", None)
    if database is not None:
        address = getattr(database, "address", None)
        if address is not None:
            hostname, port = address

    request = Request(
        command=ensure_utf8(command),
        key=ensure_utf8(key),
        hostname=ensure_utf8(hostname),
        port=ensure_utf8(port),
        connectionName=None,
        db=ensure_utf8(getattr(database, "name", None)),
        table=ensure_utf8(getattr(instance, "name", None)),
    )
    request = request._asdict()
    context.iopipe.mark.db_trace(trace, "mongodb", request)


def collect_redis_metrics(context, trace, args, connection):
    command, key = None, None
    if len(args) > 0:
        command = args[0]
    if len(args) > 1:
        key = args[1]

    request = Request(
        command=ensure_utf8(command),
        key=ensure_utf8(key),
        hostname=ensure_utf8(connection.get("host", "localhost")),
        port=ensure_utf8(connection.get("port", 6379)),
        connectionName=None,
        db=ensure_utf8(connection.get("db", 0)),
        table=None,
    )
    request = request._asdict()
    context.iopipe.mark.db_trace(trace, "redis", request)


def patch_pymongo(context):
    """
    Monkey patches pymongo client, if available. Overloads the
    query methods to add tracing and metrics collection.
    """

    def wrapper(wrapped, instance, args, kwargs):
        if not hasattr(context, "iopipe") or not hasattr(
            context.iopipe, "mark"
        ):  # pragma: no cover
            return wrapped(*args, **kwargs)

        id = ensure_utf8(str(uuid.uuid4()))
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_pymongo_metrics(context, trace, instance, response)
        return response

    try:
        wrapt.wrap_function_wrapper("pymongo.collection", "Collection.find", wrapper)
    except Exception:  # pragma: no cover
        pass
    else:
        for class_method in (
            "bulk_write",
            "delete_many",
            "delete_one",
            "insert_many",
            "insert_one",
            "replace_one",
            "update_many",
            "update_one",
        ):
            wrapt.wrap_function_wrapper(
                "pymongo.collection", "Collection.%s" % class_method, wrapper
            )


def patch_redis(context):
    """
    Monkey patches redis client, if available. Overloads the
    execute methods to add tracing and metrics collection.
    """

    def wrapper(wrapped, instance, args, kwargs):
        if not hasattr(context, "iopipe") or not hasattr(
            context.iopipe, "mark"
        ):  # pragma: no cover
            return wrapped(*args, **kwargs)

        id = ensure_utf8(str(uuid.uuid4()))
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_redis_metrics(
            context, trace, args, instance.connection_pool.connection_kwargs
        )
        return response

    def pipeline_wrapper(wrapped, instance, args, kwargs):  # pragma: no cover
        if not hasattr(context, "iopipe") or not hasattr(
            context.iopipe, "mark"
        ):  # pragma: no cover
            return wrapped(*args, **kwargs)

        # We don't need the entire command stack, just collect a stack count
        pipeline_args = ("PIPELINE", ensure_utf8(len(instance.command_stack)))

        id = ensure_utf8(str(uuid.uuid4()))
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_redis_metrics(
            context, trace, pipeline_args, instance.connection_pool.connection_kwargs
        )
        return response

    try:
        wrapt.wrap_function_wrapper("redis.client", "Redis.execute_command", wrapper)
    except Exception:  # pragma: no cover
        pass
    else:
        for module_name, class_method in [
            ("redis.client", "Pipeline.execute"),
            ("redis.client", "Pipeline.immediate_execute_command"),
        ]:
            wrapt.wrap_function_wrapper(module_name, class_method, pipeline_wrapper)


def restore_pymongo():
    """Restores pymongo"""
    try:
        from pymongo.collection import Collection
    except ImportError:  # pragma: no cover
        pass
    else:
        for class_method in (
            "bulk_write",
            "delete_many",
            "delete_one",
            "insert_many",
            "insert_one",
            "replace_one",
            "update_many",
            "update_one",
        ):
            if hasattr(getattr(Collection, class_method), "__wrapped__"):
                setattr(
                    Collection,
                    class_method,
                    getattr(Collection, class_method).__wrapped__,
                )


def restore_redis():
    """Restores the redis client"""
    try:
        from redis.client import Pipeline, Redis
    except ImportError:  # pragma: no cover
        pass
    else:
        if hasattr(Redis.execute_command, "__wrapped__"):
            setattr(Redis, "execute_command", Redis.execute_command.__wrapped__)
        for class_method in ["execute", "immediate_execute_command"]:
            if hasattr(getattr(Pipeline, class_method), "__wrapped__"):
                setattr(
                    Pipeline, class_method, getattr(Pipeline, class_method).__wrapped__
                )


def patch_db_requests(context):
    if not hasattr(context, "iopipe"):
        return

    patch_pymongo(context)
    patch_redis(context)


def restore_db_requests():
    restore_pymongo()
    restore_redis()



import imp
import os

from iopipe import IOpipe

iopipe = IOpipe()
wrapped_handler = None


@iopipe
def wrapper(event, context):
    return get_wrapped_handler()(event, context)


def get_handler():
    if "IOPIPE_HANDLER" not in os.environ or not os.environ["IOPIPE_HANDLER"]:
        raise ValueError("No value specified in IOPIPE_HANDLER environment variable")

    try:
        module_path, handler_name = os.environ["IOPIPE_HANDLER"].rsplit(".", 1)
    except ValueError:
        raise ValueError(
            "Improperly formated handler value: %s" % os.environ["IOPIPE_HANDLER"]
        )

    module_path = module_path.replace("/", ".")
    file_handle, pathname, desc = None, None, None

    try:
        for segment in module_path.split("."):
            if pathname is not None:
                pathname = [pathname]

            file_handle, pathname, desc = imp.find_module(segment, pathname)

        if file_handle is None:
            module_type = desc[2]
            if module_type == imp.C_BUILTIN:
                raise ImportError(
                    "Cannot use built-in module %s as a handler module" % module_path
                )

        module = imp.load_module(module_path, file_handle, pathname, desc)
    except Exception as e:
        raise ImportError("Failed to import module: %s" % module_path)
    finally:
        if file_handle is not None:
            file_handle.close()

    try:
        handler = getattr(module, handler_name)
    except AttributeError as e:
        raise AttributeError("No handler %s in module %s" % (handler_name, module_path))

    return handler


def get_wrapped_handler():
    global wrapped_handler

    if not wrapped_handler:
        wrapped_handler = get_handler()

    return wrapped_handler



import logging
import os
import time

try:
    import requests
except ImportError:
    from botocore.vendored import requests

from .collector import SUPPORTED_REGIONS

logger = logging.getLogger(__name__)


def get_signer_hostname():
    """
    Returns the IOpipe signer hostname for a region

    :returns: The signer hostname
    :rtype str
    """
    region = os.getenv("AWS_REGION", "")
    region = region if region and region in SUPPORTED_REGIONS else "us-west-2"
    return "signer.{region}.iopipe.com".format(region=region)


def get_signed_request(config, context, extension):
    """
    Returns a signed request URL from IOpipe

    :param config: The IOpipe config
    :param context: The AWS context to request a signed URL
    :param extension: The extension of the file to sign
    :returns: A signed request URL
    :rtype: str
    """
    url = "https://{hostname}/".format(hostname=get_signer_hostname())

    try:
        logger.debug("Requesting signed request URL from %s", url)
        response = requests.post(
            url,
            json={
                "arn": context.invoked_function_arn,
                "requestId": context.aws_request_id,
                "timestamp": int(time.time() * 1000),
                "extension": extension,
            },
            headers={"Authorization": config["token"]},
            timeout=config["network_timeout"],
        )
        response.raise_for_status()
    except Exception as e:
        logger.debug("Error requesting signed request URL: %s", e)
        if hasattr(e, "response"):
            logger.debug(e.response.content)
    else:
        response = response.json()
        logger.debug("Signed request URL received for %s", response["url"])
        return response



import collections
import uuid
import wrapt

from iopipe.compat import string_types, urlparse
from .util import ensure_utf8

INCLUDE_HEADERS = [
    "content-length",
    "content-type",
    "host",
    "server",
    "user-agent",
    "x-amz-target",
]

Request = collections.namedtuple(
    "Request",
    [
        "hash",
        "headers",
        "hostname",
        "method",
        "path",
        "pathname",
        "port",
        "protocol",
        "query",
        "url",
    ],
)

Response = collections.namedtuple(
    "Response", ["headers", "statusCode", "statusMessage"]
)


def patch_requests_session_send(context, http_filter, http_headers):
    """
    Monkey patches requests' session class, if available. Overloads the
    send method to add tracing and metrics collection.
    """

    def wrapper(wrapped, instance, args, kwargs):
        if not hasattr(context, "iopipe") or not hasattr(context.iopipe, "mark"):
            return wrapped(*args, **kwargs)
        id = ensure_utf8(str(uuid.uuid4()))
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_metrics_for_response(
            response.request, response, context, trace, http_filter, http_headers
        )
        return response

    try:
        wrapt.wrap_function_wrapper("requests.sessions", "Session.send", wrapper)
    except Exception:  # pragma: no cover
        pass


def patch_botocore_session_send(context, http_filter, http_headers):
    """
    Monkey patches botocore's session, if available. Overloads the
    session class' send method to add tracing and metric collection.
    """

    def wrapper(wrapped, instance, args, kwargs):
        if not hasattr(context, "iopipe") or not hasattr(context.iopipe, "mark"):
            return wrapped(*args, **kwargs)
        id = str(uuid.uuid4())
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_metrics_for_response(
            args[0], response, context, trace, http_filter, http_headers
        )
        return response

    try:
        wrapt.wrap_function_wrapper(
            "botocore.httpsession", "URLLib3Session.send", wrapper
        )
    except Exception:  # pragma: no cover
        pass


def patch_botocore_vendored_session_send(context, http_filter, http_headers):
    """
    Monkey patches botocore's vendored requests, if available. Overloads the
    session class' send method to add tracing and metric collection.
    """

    def wrapper(wrapped, instance, args, kwargs):
        if not hasattr(context, "iopipe") or not hasattr(context.iopipe, "mark"):
            return wrapped(*args, **kwargs)
        id = str(uuid.uuid4())
        with context.iopipe.mark(id):
            response = wrapped(*args, **kwargs)
        trace = context.iopipe.mark.measure(id)
        context.iopipe.mark.delete(id)
        collect_metrics_for_response(
            response.request, response, context, trace, http_filter, http_headers
        )
        return response

    try:
        wrapt.wrap_function_wrapper(
            "botocore.vendored.requests.sessions", "Session.send", wrapper
        )
    except Exception:  # pragma: no cover
        pass


def restore_requests_session_send():
    """Restores the original requests session send method"""
    try:
        from requests.sessions import Session as RequestsSession
    except ImportError:  # pragma: no cover
        pass
    else:
        if hasattr(RequestsSession.send, "__wrapped__"):
            setattr(RequestsSession, "send", RequestsSession.send.__wrapped__)


def restore_botocore_session_send():
    """Restores the original botocore session send method"""
    try:
        from botocore.httpsession import URLLib3Session as BotocoreSession
    except ImportError:  # pragma: no cover
        pass
    else:
        if hasattr(BotocoreSession.send, "__wrapped__"):
            setattr(BotocoreSession, "send", BotocoreSession.send.__wrapped__)


def restore_botocore_vendored_session_send():
    """Restores the original botocore vendored session send method"""
    try:
        from botocore.vendored.requests.sessions import (
            Session as BotocoreVendoredSession,
        )
    except ImportError:  # pragma: no cover
        pass
    else:
        if hasattr(BotocoreVendoredSession.send, "__wrapped__"):
            setattr(
                BotocoreVendoredSession,
                "send",
                BotocoreVendoredSession.send.__wrapped__,
            )


def patch_http_requests(context, http_filter, http_headers):
    if not hasattr(context, "iopipe"):
        return

    patch_requests_session_send(context, http_filter, http_headers)
    patch_botocore_session_send(context, http_filter, http_headers)
    patch_botocore_vendored_session_send(context, http_filter, http_headers)


def restore_http_requests():
    restore_requests_session_send()
    restore_botocore_session_send()
    restore_botocore_vendored_session_send()


def collect_metrics_for_response(
    http_request, http_response, context, trace, http_filter, http_headers
):
    """
    Collects relevant metrics from a requests Response object and adds them to
    the IOpipe context.
    """
    include_headers = INCLUDE_HEADERS
    if isinstance(http_headers, (list, tuple)):
        include_headers = include_headers + [
            key.lower()
            for key in http_headers
            if isinstance(http_headers, string_types)
        ]

    request = None
    if http_request:
        parsed_url = None
        if hasattr(http_request, "url"):
            parsed_url = urlparse(http_request.url)

        request_headers = []
        if hasattr(http_request, "headers"):
            request_headers = [
                {"key": ensure_utf8(k), "string": ensure_utf8(v)}
                for k, v in http_request.headers.items()
                if k.lower() in include_headers
            ]

        request = Request(
            hash=ensure_utf8(getattr(parsed_url, "fragment", None)),
            headers=request_headers,
            hostname=ensure_utf8(getattr(parsed_url, "hostname", None)),
            method=ensure_utf8(getattr(http_request, "method", None)),
            path=ensure_utf8(getattr(parsed_url, "path", None)),
            # TODO: Determine if this is redundant
            pathname=ensure_utf8(getattr(parsed_url, "path", None)),
            port=ensure_utf8(getattr(parsed_url, "port", None)),
            protocol=ensure_utf8(getattr(parsed_url, "scheme", None)),
            query=ensure_utf8(getattr(parsed_url, "query", None)),
            url=ensure_utf8(getattr(http_request, "url", None)),
        )

        # TODO: Possibly remove the namedtuple in favor of just a dict
        request = request._asdict()

    response_headers = []
    if hasattr(http_response, "headers"):
        response_headers = [
            {"key": ensure_utf8(k), "string": ensure_utf8(v)}
            for k, v in http_response.headers.items()
            if k.lower() in include_headers
        ]

    response = Response(
        headers=response_headers,
        statusCode=ensure_utf8(getattr(http_response, "status_code", None)),
        statusMessage=None,
    )

    # TODO: Possibly remove the namedtuple in favor of just a dict
    response = response._asdict()

    if http_filter is not None and callable(http_filter):
        try:
            request, response = http_filter(request, response)
        except Exception:
            return

    context.iopipe.mark.http_trace(trace, request, response)



from .response_types import LambdaProxy
from .util import collect_all_keys, get_value, has_key, slugify


class EventType(object):
    keys = []
    response_keys = []
    exclude_keys = []
    required_keys = []
    response_type = None
    source = None

    def __init__(self, event):
        self.event = event
        if self.response_type:
            self.response_type = self.response_type(self.type)

    @property
    def slug(self):
        return slugify(self.type)

    def has_required_keys(self):
        return all(has_key(self.event, key) for key in self.required_keys)

    def collect(self):
        if self.keys == "all":
            event_info = collect_all_keys(
                self.event, "@iopipe/event-info.%s" % self.type, self.exclude_keys
            )
            event_info["@iopipe/event-info.eventType"] = self.type
            return event_info

        event_info = {}
        for key in self.keys:
            if isinstance(key, tuple):
                old_key, new_key = key
            else:
                old_key = new_key = key
            value = get_value(self.event, old_key)
            if value is not None:
                event_info["@iopipe/event-info.%s.%s" % (self.type, new_key)] = value
        event_info["@iopipe/event-info.eventType"] = self.type
        if self.source:
            event_info["@iopipe/event-info.eventType.source"] = self.source
        return event_info


class ALB(EventType):
    type = "alb"
    keys = [
        "headers.host",
        'headers.["user-agent"]',
        'headers.["x-amzn-trace-id"]',
        "httpMethod",
        "path",
        "requestContext.elb.targetGroupArn",
    ]
    required_keys = ["httpMethod", "requestContext.elb"]


class AlexaSkill(EventType):
    type = "alexaSkill"
    keys = "all"
    exclude_keys = [
        "context.System.apiAccessToken",
        "context.System.user.accessToken",
        "context.System.user.permissions.consentToken",
    ]
    required_keys = [
        "context.System",
        "request.requestId",
        "session.attributes",
        "session.user",
    ]


class ApiGateway(EventType):
    type = "apiGateway"
    keys = [
        "httpMethod",
        "path",
        "requestContext.accountId",
        "requestContext.httpMethod",
        "requestContext.identity.userAgent",
        "requestContext.requestId",
        "requestContext.resourcePath",
        "requestContext.stage",
        "resource",
    ]
    required_keys = ["headers", "httpMethod", "path", "requestContext", "resource"]
    response_type = LambdaProxy


class CloudFront(EventType):
    type = "cloudFront"
    keys = [
        "Records[0].cf.config.distributionId",
        "Records[0].cf.request.clientIp",
        "Records[0].cf.request.headers.host[0].value",
        'Records[0].cf.request.headers.["user-agent"][0].value',
        "Records[0].cf.request.method",
        "Records[0].cf.request.uri",
    ]
    required_keys = ["Records[0].cf"]


class Firehose(EventType):
    type = "firehose"
    keys = ["deliveryStreamArn", "region"]
    required_keys = [
        "deliveryStreamArn",
        "records[0]",
        "records[0].kinesisRecordMetadata",
    ]


class Kinesis(EventType):
    type = "kinesis"
    keys = ["Records.length", "Records[0].awsRegion", "Records[0].eventSourceARN"]
    required_keys = ["Records[0].eventVersion", "Records[0].eventSource"]

    def has_required_keys(self):
        return (
            super(Kinesis, self).has_required_keys()
            and get_value(self.event, "Records[0].eventVersion") == "1.0"
            and get_value(self.event, "Records[0].eventSource") == "aws:kinesis"
        )


class S3(EventType):
    type = "s3"
    keys = [
        "Records[0].awsRegion",
        "Records[0].eventName",
        "Records[0].eventTime",
        "Records[0].requestParameters.sourceIPAddress",
        'Records[0].responseElements["x-amz-id-2"]',
        'Records[0].responseElements["x-amz-request-id"]',
        "Records[0].s3.bucket.arn",
        "Records[0].s3.bucket.name",
        "Records[0].s3.object.key",
        "Records[0].s3.object.sequencer",
        "Records[0].s3.object.size",
        "Records[0].userIdentity.principalId",
    ]
    required_keys = ["Records[0].eventVersion", "Records[0].eventSource"]

    def has_required_keys(self):
        return (
            super(S3, self).has_required_keys()
            and get_value(self.event, "Records[0].eventVersion") in ("2.0", "2.1")
            and get_value(self.event, "Records[0].eventSource") == "aws:s3"
        )


class Scheduled(EventType):
    type = "scheduled"
    keys = ["account", "id", "region", "resources[0]", "time"]
    required_keys = ["source"]

    def has_required_keys(self):
        return (
            super(Scheduled, self).has_required_keys()
            and get_value(self.event, "source") == "aws.events"
        )


class ServerlessLambda(EventType):
    type = "apiGateway"
    source = "slsIntegrationLambda"
    keys = [
        ('headers.["X-Amz-Cf-Id"]', "headers.X-Amz-Cf-Id"),
        ('headers.["X-Amzn-Trace-Id"]', "headers.X-Amzn-Trace-Id"),
        ("identity.accountId", "requestContext.accountId"),
        ("identity.userAgent", "requestContext.identity.userAgent"),
        ("method", "httpMethod"),
        ("method", "requestContext.httpMethod"),
        ("stage", "requestContext.stage"),
    ]
    required_keys = ["identity.userAgent", "identity.sourceIp", "identity.accountId"]
    response_type = LambdaProxy


class SES(EventType):
    type = "ses"
    keys = [
        "Records[0].ses.mail.commonHeaders.date",
        "Records[0].ses.mail.commonHeaders.messageId",
        "Records[0].ses.mail.commonHeaders.returnPath",
        "Records[0].ses.mail.commonHeaders.subject",
        "Records[0].ses.mail.messageId",
        "Records[0].ses.mail.source",
    ]
    required_keys = ["Records[0].eventVersion", "Records[0].eventSource"]

    def has_required_keys(self):
        return (
            super(SES, self).has_required_keys()
            and get_value(self.event, "Records[0].eventVersion") == "1.0"
            and get_value(self.event, "Records[0].eventSource") == "aws:ses"
        )


class SNS(EventType):
    type = "sns"
    keys = [
        "Records[0].EventSubscriptionArn",
        "Records[0].Sns.MessageId",
        "Records[0].Sns.Signature",
        "Records[0].Sns.SignatureVersion",
        "Records[0].Sns.SigningCertUrl",
        "Records[0].Sns.UnsubscribeUrl",
        "Records[0].Sns.Subject",
        "Records[0].Sns.Timestamp",
        "Records[0].Sns.TopicArn",
        "Records[0].Sns.Type",
    ]
    required_keys = ["Records[0].EventVersion", "Records[0].EventSource"]

    def has_required_keys(self):
        return (
            super(SNS, self).has_required_keys()
            and get_value(self.event, "Records[0].EventVersion") == "1.0"
            and get_value(self.event, "Records[0].EventSource") == "aws:sns"
        )


class SQS(EventType):
    type = "sqs"
    keys = [
        "Records[0].attributes.ApproximateFirstReceiveTimestamp",
        "Records[0].attributes.ApproximateReceiveCount",
        "Records[0].attributes.SenderId",
        "Records[0].attributes.SentTimestamp",
        "Records[0].awsRegion",
        "Records[0].eventSourceARN",
        "Records[0].md5OfBody",
        "Records[0].messageId",
        "Records[0].receiptHandle",
    ]
    required_keys = ["Records[0].eventSource"]

    def has_required_keys(self):
        return (
            super(SQS, self).has_required_keys()
            and get_value(self.event, "Records[0].eventSource") == "aws:sqs"
        )


EVENT_TYPES = [
    ALB,
    AlexaSkill,
    ApiGateway,
    CloudFront,
    Firehose,
    Kinesis,
    S3,
    SES,
    SNS,
    SQS,
    Scheduled,
    ServerlessLambda,
]


def metrics_for_event_type(event, context):
    for EventType in EVENT_TYPES:
        event_type = EventType(event)
        if event_type.has_required_keys():
            context.iopipe.label("@iopipe/plugin-event-info")
            context.iopipe.label("@iopipe/aws-%s" % event_type.slug)
            context.iopipe.event_type("aws-%s" % event_type.slug)
            event_info = event_type.collect()
            [context.iopipe.metric(k, v) for k, v in event_info.items()]
            break

    if context.iopipe.is_step_function:
        context.iopipe.collect_step_meta(event)
        if context.iopipe.step_meta:
            for key, value in context.iopipe.step_meta.items():
                context.iopipe.metric("@iopipe/event-info.stepFunction.%s" % key, value)

    return event_type


def metrics_for_response_type(event_type, context, response):
    if event_type.response_type:
        response_info = event_type.response_type.collect(response)
        [context.iopipe.metric(k, v) for k, v in response_info.items()]



import jmespath
import re

# Caches previously seen parsers
parsers = {}


def _format_path(path):
    return path.replace('["', '."').replace('.."', '."').replace('"]', '"')


def _get_parser(path):
    path = _format_path(path)
    parser = parsers.get(path)
    if parser is None:
        parser = jmespath.compile(path)
        parsers[path] = parser
    return parser


def get_value(obj, path):
    if path.endswith(".length"):
        path = path.replace(".length", "|length(@)")
    parser = _get_parser(path)
    return parser.search(obj)


def has_key(obj, path):
    parser = _get_parser(path)
    result = parser.search(obj)
    return result is not None


def collect_all_keys(obj, initial_path, exclude_keys=None):
    out = {}

    def flatten(o, path=None):
        if path is None:
            path = [initial_path]
        if isinstance(o, dict):
            for key in o:
                flatten(o[key], path + [key])
        elif isinstance(o, list):
            for i, value in enumerate(o):
                flatten(value, path + [str(i)])
        else:
            if exclude_keys and ".".join(path[1:]) in exclude_keys:
                return
            out[".".join(path)] = o

    flatten(obj)
    return out


def slugify(name):
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1-\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1-\2", s1).lower()



import os

from posixpath import join as urljoin

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

SUPPORTED_REGIONS = [
    "ap-northeast-1",
    "ap-northeast-2",
    "ap-south-1",
    "ap-southeast-1",
    "ap-southeast-2",
    "ca-central-1",
    "eu-central-1",
    "eu-west-1",
    "eu-west-2",
    "us-east-2",
    "us-west-1",
    "us-west-2",
]


def get_collector_path(base_url=None):
    """
    Returns the IOpipe collector's path. By default this is `/v0/event`.

    :param base_url: An optional base URL to use.
    :returns: The collector's path.
    :rtype: str
    """
    if not base_url:
        return "/v0/event"
    event_url = urlparse(base_url)
    event_path = urljoin(event_url.path, "v0/event")
    if not event_path.startswith("/"):
        event_path = "/%s" % event_path
    if event_url.query:
        event_path = "?".join([event_path, event_url.query])
    return event_path


def get_hostname(config_url=None):
    """
    Returns the IOpipe collector's hostname. If  the `AWS_REGION` environment
    variable is not set or unsupported then `us-east-1` will be used by
    default. In this case, `us-east-1` is `metrics-api.iopipe.com`.

    :param config_url: A optional config URL to use.
    :returns: The collector's hostname.
    :rtype: str
    """
    region_string = ""
    if config_url:
        return urlparse(config_url).hostname
    aws_region = os.getenv("AWS_REGION")
    if aws_region and aws_region in SUPPORTED_REGIONS:
        region_string = ".%s" % aws_region
    return "metrics-api%s.iopipe.com" % region_string
